
#use "recup_donnees.ml";;

(* les types utiliss pour manipuler des graphes *)

type color = int;;

type node = { etiquette:int; mutable couleur:color; mutable nb_voisins:int};;

type graphe = { vert: node list; edges: (int * int) list};;




let affiche_dims tab = print_string "nb ligne = ";
(* affiche les dimensions d'un tableau  deux dimensions *)
  print_int (Array.length tab);
  print_string " et nb col = "; print_int (Array.length (tab.(0)));
  print_string "\n";;


let rec separe_aretes sommet = function
(* coupe les artes en deux tas suivants qu'elles sont adjacentes
    sommet ou non *)
    [] -> ([], [])
  | a::b -> let (u, v) = separe_aretes sommet b in
      match a with
	  (c, d) -> if (d = sommet or c = sommet) then (a::u, v)
	  else (u, a::v);;

let rec compte_aretes sommet = function
(* calcule le degree d'un sommet *)
    [] -> 0
  | a::b -> let u = compte_aretes sommet b in
    match a with
      (c, d) -> if (c = sommet.etiquette or d = sommet.etiquette) then (u + 1) else u;;

let rec coupe_sommet pivot = function
(* la fonction coupe du tri rapide suivant le degree *)
    [] -> ([], [])
  | a::b -> let (u, v) = coupe_sommet pivot b in
    if (a.nb_voisins < pivot) then (a::u, v) else (u, a::v);;

let rec trie_sommet = function
(* le tri rapide *)
  | [] -> []
  | a::b -> let (u, v) = coupe_sommet a.nb_voisins b in
    (trie_sommet u)@[a]@(trie_sommet v);;

let rec calcule_nb_voisins = function
(* met  jour le champ nb_voisin dans la structure des sommets *)
    [], l -> ()
  | a::b, l -> (a.nb_voisins <- (compte_aretes a l); calcule_nb_voisins (b, l));;

let rec extract_neighbours sommet = function
(* rcupre la liste des voisins d'un sommet  partir de la liste
   de ses artes adjacentes *)
    [] -> []
  | a::b -> match a with 
	(c, d) -> if (c = sommet) then d::(extract_neighbours sommet b)
	else c::(extract_neighbours sommet b);;

let rec pick_vert_color sommet = function
(* rcupre la liste des couleurs d'un sommets *)
  | [] -> failwith "Sommet inexistant dans pick_vert_color"
  | a::b -> if (a.etiquette = sommet) then a.couleur else pick_vert_color sommet b;;

let rec gather_colors ve = function
(* rcupre la liste des couleurs d'une liste de sommets *)
  | [] -> []
  | a::b -> (pick_vert_color a ve)::(gather_colors ve b);;

let rec elimine_couleur a = function
(* enleve une couleur d'une liste de couleurs *)
  | [] -> []
  | b::c -> if (b = a) then (elimine_couleur a c) else b::(elimine_couleur a c);;

let rec soustract_colors = function
(* diffrence ensembliste de deux ensembles de couleurs *)
  | [], l -> l
  | l, [] -> []
  | a::b, l -> soustract_colors (b, (elimine_couleur a l));;

let rec elimine_doublon = function
(* comme son nom l'indique *)
  | [] -> []
  | a::b -> a::(elimine_couleur a (elimine_doublon b));;      

let rec max_color = function
(* le numero maximal des couleurs utiliss,
   ie le nombre de couleurs utilises *)
  | [] -> 0
  | a::b -> let u = max_color b in if (a > u) then a else u;;

let rec colorie_graphe g =
(* colorie le graphe suivant l'algo rcursif du Grne *)
  calcule_nb_voisins (g.vert, g.edges);
  let vert_s = trie_sommet g.vert in
    (match vert_s with
	 [] -> (g.vert, [])
       | a::b -> let (u, v) = separe_aretes a.etiquette g.edges in
		 let (w, x) = colorie_graphe {vert = b; edges = v} in
		 let k = gather_colors w (extract_neighbours a.etiquette u)  in
		 let avail = soustract_colors (k, x) in
		   (match avail with
			[] -> let z = (max_color x) + 1 in
			  a.couleur <- z; (a::w, z::x)
		      | h::j -> a.couleur <- h; (a::w, x)
		   )
    );;



let make_node i = {etiquette = i; couleur = 0; nb_voisins = 0};;
(* pour crer facilement des graphes *)

let rec make_several_nodes = function
(* pour crr facilement des graphes *)
  | -1 -> []
  | n -> (make_node n)::(make_several_nodes (n - 1));;


let calcule_graphe tab m n = 
(* du tableau de transitions non comprss  son graphe
   d'incompatibilits *)
  let tmped = ref [] in
  let verti = make_several_nodes (m - 1) in
    for i = 0 to m - 1 do
      for j = i + 1 to m - 1 do
	let da_flag = ref true in
	let k = ref 0 in
	  while (!da_flag) do
	    if ((tab.(i).(!k) <> tab.(j).(!k)) && (tab.(i).(!k) >= 0) && (tab.(j).(!k) >= 0))
	    then (tmped := (i,j)::(!tmped); 
		  da_flag := false;);
	    k := !k + 1;
	    if (!k = n) then da_flag := false;
	  done;
      done;
    done;
    { vert = verti; edges = !tmped};;

let rec color_part col = function
(* rcupre les sommets d'une couleur donne dans une liste 
   de sommets *)
    [] -> []
  | a::b -> if (a.couleur = col) then (a.etiquette)::(color_part col b) else color_part col b;;

let genere_nouveau_tab tab_ini m n verti delta =
(* le nouveau tableau de transitions, comprss,  partir
   du graphe et du tableau initial. Gnre aussi le tableau de couleur
   des lignes *)
  let the_matrix_strikes_back = Array.make_matrix delta n 0 in
    for i = 1 to delta do
      let rec aux = function
	  [] -> ()
	| a::b -> 
	    for j = 0 to n - 1 do
	      if (tab_ini.(a).(j) >= 0) then the_matrix_strikes_back.(i - 1).(j) <- tab_ini.(a).(j);
	    done;
	    aux b;
      in
	aux (color_part i verti);
    done;

    let new_lines = Array.make m 0 in
    let rec aux_prime = function
	[] -> ()
      | a::b -> new_lines.(a.etiquette) <- (a.couleur - 1); aux_prime b;
    in
      aux_prime verti;
      (the_matrix_strikes_back, new_lines);;

let gen_bitmap tab_ini m n = 
(* le bitmap des transitions rates, description dtaille dans le bout
   de code C *)
  let da_bitmap = Array.make_matrix m  (n / 8) 0 in
    for i = 0 to m - 1 do
      for j = 0 to (n/8) - 1 do
	let c = ref 0 in
	  for k = (8 * j) to (8 * j) + 7 do
	    if (tab_ini.(i).(k) < 0) then 
	      c := (!c * 2) + 1
	    else
	      c := (!c * 2);
	  done;
	  da_bitmap.(i).(j) <- !c;
      done;
    done;
    da_bitmap;;

let ecrit_yy_next_with_graph tab m n sortie =
(* ecrit la nouvelle table de transitions dans le nouveau lex.yy.c *)
  affiche_dims tab;
  output_string sortie "yy_next_with_graph[][128] =\n{";
  for i = 0 to m - 1 do
    output_string sortie "{\n";
    for j = 0 to n - 2 do
      output_string sortie (string_of_int (tab.(i).(j)));
      output_string sortie ", ";
    done;
    output_string sortie (string_of_int (tab.(i).(n - 1)));
    output_string sortie "}";
    if (i <> (m - 1)) then output_string sortie ",\n";
  done;
  output_string sortie "};\n";;

let ecrit_yy_old_state_to_line tab m sortie = 
(* ecrit la table de couleurs dans le nouveau lex.yy.c *)
  output_string sortie "short yy_old_state_to_line[] =\n{";
  for i = 0 to m - 1 do
    output_string sortie (string_of_int (tab.(i)));
    if (i <> (m - 1)) then output_string sortie ", ";
  done;
  output_string sortie "};\n";;

let ecrit_yy_bitmap_graph tab m sortie = 
(* ecrit la table de bitmaps dans le nouveau lex.yy.c *)
  output_string sortie "char yy_bitmap_graph[][16] = \n{\n";
  let n = 16 in
  for i = 0 to m - 1 do
    output_string sortie "{\n";
    for j = 0 to n - 2 do
      output_string sortie (string_of_int tab.(i).(j));
      output_string sortie ", ";
    done;
    output_string sortie (string_of_int tab.(i).(n - 1));
    output_string sortie "}";
    if (i <> (m - 1)) then output_string sortie ",\n";
  done;
  output_string sortie "};\n";;


let gros_bout_de_c = 
(* les fonctions de manipulations des nouveaux tableaux *)
  "
/*
  Le programme Ocaml a cree correctement les tableaux:
  - yy_next_with_graph
  - yy_old_state_to_line
  - yy_bitmap_graph

  yy_next_with_graph est le tableau des lignes de transitions
  aplaties par l'algo de coloriage de graphe.

  yy_old_state_to_line permet de savoir pour un etat
  dans quelle ligne regarder dans la nouvelle table
  de transition.

  yy_bitmap_graph est un tableau de char  16 colonnes et
  avec autant de lignes que d'tats. La transition 
  (state, char) est une transition vers echec si le bit
  de rang (char % 8) de l'element de colonne (char / 8) 
  et de ligne state est  1, ce qui correspond effectivement
  au char ieme bit du tableau si on les concatenait.

*/


int bitmap_check(yy_state_type state, unsigned int lu)
{
  int q, r;
  unsigned char c;

  q = lu / 8; /* un char a 8 bits */
  r = lu % 8;
  c = yy_bitmap_graph[state][q];

 if (c & (1 << (7 - r)))
   return 0;
 else
   return 1;
}

yy_state_type yy_nxt(yy_state_type state, unsigned int lu)
{
  if (bitmap_check(state, lu))
    return yy_next_with_graph[yy_old_state_to_line[state]][lu];
  else
    return - state;
}
";;



let laissons_passer_jusque_yy_nxt entree sortie = 
(* il faut laisser passer jusque yy_nxt en l'enlevant *)
  let aux  = function
      (0, 'y') -> 1
    | (1, 'y') -> 2
    | (2, '_') -> 3
    | (3, 'n') -> 4
    | (4, 'x') -> 5
    | (5, 't') -> 6
    | (2, 'y') -> 1
    | (_, _) -> 0
  in
  let matched_aux = function
    | 0 -> ""
    | 1 -> "y"
    | 2 -> "yy"
    | 3 -> "yy_"
    | 4 -> "yy_n"
    | 5 -> "yy_nx"
    | _ -> failwith "erreur dans matched_aux"
  in
  let q = ref 0 in
  while (!q <> 6) do
    let last_state = !q in
    let c = input_char entree in
    q := aux (!q, c);
    if (!q = 0) then
      (
       output_string sortie (matched_aux last_state);
       output_char sortie c;
      );
  done;;

let recuperer_truc_entre_crochets entree =
(* pour changer facilement les occurences suivantes de yy_nxt
   il faut remplacer les acces au tableau par un appel de fonction *)
  while ((input_char entree) <> '[') do
    ()
  done;
  let buffer = ref "" in
  let k = ref true in
  while (!k) do
    let c = input_char entree in
    if (c = ']') then
      k := false
    else
      (
       buffer := !buffer ^ "e";
       String.set !buffer ((String.length !buffer) - 1) c;
      )
    done;
  !buffer;;
      
  
let manger_jusque_semi_colomn entree = 
(* pour effacer l'ancien tableau de transitions non compresse *)
  while ((input_char entree) <> ';') do
    ()
  done;;

let main_graphe() =
(* la fonction principale *)
  let filename = Sys.argv.(2) in 
  let tab = get_yy_nxt Sys.argv.(1) in
  
    print_string ("FICHIER DE TRAVAIL : "^(Sys.argv.(1)));  
    print_string "Au depart le tableau a les dimensions:\n";
    affiche_dims tab;
    
    let m = Array.length tab in
    let n = Array.length (tab.(0)) in
      
      print_string "Au depart le tableau de transitions prend :\n";
      print_int (m * n);
      print_string " x sizeof(short)\n";
      print_string "soit avec sizeof(short) = 2 :\n";
      print_int (m * n * 2);
      print_string " octets.\n";
      
      let fichier = open_out filename in
	
	
      let (gra,x) = colorie_graphe (calcule_graphe tab m n) in
      let delta = max_color x in
	
	print_newline();
	print_string "VERSION 2 - GRAPH COLORING \n";
	print_string "Le graphe a "; print_int delta; print_string " couleurs.\n";
	
	let (yy_next, yy_new) = genere_nouveau_tab tab m n gra delta in
	let yy_bitmap = gen_bitmap tab m n in
	  
	let entree = open_in Sys.argv.(1) in
	  
	  laissons_passer_jusque_yy_nxt entree fichier;
	  manger_jusque_semi_colomn entree;  
	  
	  ecrit_yy_next_with_graph yy_next delta n fichier;
	  ecrit_yy_old_state_to_line yy_new m fichier;
	  ecrit_yy_bitmap_graph yy_bitmap m fichier;
	  output_string fichier gros_bout_de_c;
	  
	  try
	    (
	      while true do
		laissons_passer_jusque_yy_nxt entree fichier;
		let c1 = recuperer_truc_entre_crochets entree in
		let c2 = recuperer_truc_entre_crochets entree in
		  output_string fichier ("yy_nxt(" ^ c1 ^ ", " ^ c2 ^ ")");
	      done
	    )
	  with
	      End_of_file -> ();
		
      (* quelques statistiques pour finir *)
		
		close_out fichier;
		close_in entree;
		
		print_string "Les nouvelles tables prennent \n";
		print_int (delta * 128 );
		print_string " x sizeof(short) pour les transitions.\n";
		print_int (n / 8 * m);
		print_string " x sizeof(char) pour le bitmap.\n";
		print_int m;
		print_string " x sizeof(short) pour savoir quel etat est sur quelle ligne.\n";
		print_string "Soit un total avec sizeof(short) = 2 de :\n";
		print_int ((delta * 128 + m) * 2 + (n / 8 * m));
		print_string " octets.\n";;




(* suivent des fonctions pour tester colorie_graphe *)


let graphe_bidon = { vert = make_several_nodes 5; 
		     edges = [(1,0); (0,2); (3,2); (1,3); 
			      (1,4); (1,5); (5,4)];};;

colorie_graphe graphe_bidon;;

calcule_nb_voisins (graphe_bidon.vert, graphe_bidon.edges);;

graphe_bidon;;


main_graphe();;





